"""
Palos is a Python3 module developed and used by the yfish group, http://www.yfish.org/.
It contains code related to bioinformatics projects focusing on next-generation sequencing data, 
population genetics, genome-wide association studies, pedigree genetics, etc.
"""
version='0.1.35'
from . ProcessOptions import ProcessOptions, generate_program_doc, process_options, \
    process_function_arguments, turn_option_default_dict2argument_default_dict
from . utils import PassingData, PassingDataList, dict_map, importNumericArray, \
    figureOutDelimiter, get_gene_symbol2gene_id_set, \
    FigureOutTaxID, getColName2IndexFromHeader, getListOutOfStr, runLocalCommand, \
    getGeneIDSetGivenAccVer, calGreatCircleDistance, openGzipFile
from . Genome import GeneModel

from . algorithm import pca_module
from . algorithm.PCA import PCA
from . algorithm.RBTree import RBTree, RBDict, RBTreeIter, RBList, RBNode
from . algorithm.BinarySearchTree import binary_tree
from . io.MatrixFile import MatrixFile



#!/usr/bin/env python3
"""
A child class of MatrixFile. used to describe AlignmentDepthIntervalFile in SunsetDB.
"""

import sys, os, math
import copy, numpy
from palos.ProcessOptions import  ProcessOptions
from palos import PassingData
from palos.io.MatrixFile import MatrixFile

class AlignmentDepthIntervalFile(MatrixFile):
    __doc__ = __doc__
    option_default_dict = copy.deepcopy(MatrixFile.option_default_dict)
    option_default_dict.update({
        ('minDepth', 0, float): [None, '', 1, ''],
        })
    def __init__(self, path=None, **keywords):
        MatrixFile.__init__(self, path=path, **keywords)
        
        #summary data
        self.no_of_intervals = 0
        self.interval_value_ls = []
        self.interval_length_ls = []
        self.chromosome_size = 0
        
        self.min_interval_value = None
        self.max_interval_value = None
        self.median_interval_value = None
        
        self.min_interval_length = None
        self.max_interval_length = None
        self.median_interval_length = None
        
        self.interval_ls = []
        
    def parseRow(self, row):
        """
        """
        start, stop, length, depth = row[:4]
        start = int(start)
        stop = int(stop)
        length = int(length)
        depth = float(depth)
        return PassingData(start=start, stop=stop, length=length, depth=depth)
    
    def _postProcessParsedRowDataForSummary(self, pdata=None):
        """
        2013.08.30
        """
        self.no_of_intervals += 1
        self.chromosome_size += pdata.length
        self.interval_value_ls.append(pdata.depth)
        self.interval_length_ls.append(pdata.length)
        
        if self.min_interval_value is None or pdata.depth <self.min_interval_value:
            self.min_interval_value = pdata.depth
        if self.max_interval_value is None or pdata.depth >self.max_interval_value:
            self.max_interval_value = pdata.depth
    
    def readThroughAndProvideSummary(self):
        """
        2013.08.30
            called by vervet/src/db/import/AddAlignmentDepthIntervalFile2DB.py
        """
        col_name2index= self.smartReadHeader()
        if col_name2index is None:
            pdata = self.parseRow(self._row)
            self._postProcessParsedRowDataForSummary(pdata)
        
        for row in self:
            pdata = self.parseRow(row)
            self._postProcessParsedRowDataForSummary(pdata)
        
        
        self.min_interval_length = numpy.min(self.interval_length_ls)
        self.max_interval_length = numpy.max(self.interval_length_ls)
        self.median_interval_length = numpy.median(self.interval_length_ls)
        
        self.mean_interval_value=numpy.mean(self.interval_value_ls)
        self.median_interval_value=numpy.median(self.interval_value_ls)
        return PassingData(
            no_of_intervals=self.no_of_intervals,
            chromosome_size=self.chromosome_size,
            mean_interval_value=self.mean_interval_value,
            median_interval_value=self.median_interval_value,
            min_interval_value=self.min_interval_value,
            max_interval_value=self.max_interval_value,
            
            min_interval_length=self.min_interval_length,
            max_interval_length=self.max_interval_length,
            median_interval_length=self.median_interval_length)
    
    def _postProcessParsedRowDataForInterval(self, pdata=None, minDepth=None, maxDepth=None):
        """
        """
        toInclude = True
        if minDepth is not None and pdata.depth<minDepth:
            toInclude = False
        if maxDepth is not  None and pdata.depth>maxDepth:
            toInclude = False
        if toInclude:
            if len(self.interval_ls)>0:
                previousInterval = self.interval_ls[-1]
            else:
                previousInterval = None
            if previousInterval is not None and pdata.start==previousInterval.stop+1:
                #adjacent, merge the two
                previousIntervalArea = previousInterval.length*previousInterval.depth
                currentIntervalArea = pdata.length*pdata.depth
                newLength = previousInterval.length + pdata.length
                newDepth = (previousIntervalArea+currentIntervalArea)/newLength
                previousInterval.stop = pdata.stop
                previousInterval.length = newLength
                previousInterval.depth = newDepth
                if self.max_interval_length is None or newLength<self.max_interval_length:
                    self.max_interval_length = newLength
            else:
                self.interval_ls.append(pdata)
                if self.min_interval_length is None or pdata.length<self.min_interval_length:
                    self.min_interval_length = pdata.length
        
    def getAllIntervalWithinDepthRange(self, minDepth=None, maxDepth=None):
        """
        2013.08.30
        """
        col_name2index= self.smartReadHeader()
        if col_name2index is None:
            pdata = self.parseRow(self._row)
            self._postProcessParsedRowDataForInterval(pdata, minDepth=minDepth, maxDepth=maxDepth)
        
        for row in self:
            pdata = self.parseRow(row)
            self._postProcessParsedRowDataForInterval(pdata, minDepth=minDepth, maxDepth=maxDepth)
        return self.interval_ls
        
if __name__ == '__main__':
    main_class = AlignmentDepthIntervalFile
    po = ProcessOptions(sys.argv, main_class.option_default_dict, error_doc=main_class.__doc__)
    instance = main_class(**po.long_option2value)
    instance.run()


"""
2012.12.15 data structure for AssociationPeak in PyTablesMatrixFile format 
"""
import sys, os
import csv
import tables
from tables import UInt64Col, Float64Col, StringCol
import numpy
from palos.utils import PassingData, PassingDataList
from palos.ProcessOptions import ProcessOptions
from palos.io.YHPyTables import YHTable, YHFile, castPyTablesRowIntoPassingData

class AssociationPeakTable(tables.IsDescription):
	id = UInt64Col(pos=0)
	chromosome = StringCol(64, pos=1)	#64 byte-long
	start = UInt64Col(pos=2)
	stop = UInt64Col(pos=3)
	start_locus_id = UInt64Col(pos=4)
	stop_locus_id = UInt64Col(pos=5)
	no_of_loci = UInt64Col(pos=6)
	peak_locus_id = UInt64Col(pos=7)
	peak_score = Float64Col(pos=8)

class AssociationPeakTableFile(YHFile):

	"""
	usage examples:
	
		peakFile = AssociationPeakTableFile(self.outputFname, mode='w')
		peakFile.addAttributeDict(attributeDict)
		peakFile.appendAssociationPeak(association_peak_ls=association_peak_ls)
		
		#for read-only
		peakFile = AssociationPeakTableFile(inputFname, mode='r', peakPadding=0)
		rbDict = peakFile.associationPeakRBDict
	"""
	def __init__(self, inputFname=None, mode='r', \
				tableName='association_peak', groupNamePrefix='group', tableNamePrefix='table',\
				filters=None, peakPadding=0, expectedrows=50000, autoRead=True, autoWrite=True, \
				**keywords):
		
		self.peakPadding = peakPadding
		self.associationPeakRBDict = None
		YHFile.__init__(self, path=inputFname, mode=mode, \
				tableName=tableName, groupNamePrefix=groupNamePrefix, tableNamePrefix=tableNamePrefix,\
				rowDefinition=AssociationPeakTable, filters=filters, expectedrows=expectedrows,\
				autoRead=autoRead, autoWrite=autoWrite,\
				debug=0, report=0)
		
		self.associationPeakTable = self.getTableObject(tableName=self.tableName)

	def _readInData(self, tableName=None, tableObject=None):
		"""
		2012.11.12
			similar to Stock_250kDB.constructRBDictFromResultPeak(), but from HDF5MatrixFile-like file
		"""
		YHFile._readInData(self, tableName=tableName, tableObject=tableObject)
		
		from palos.algorithm.RBTree import RBDict
		from palos.polymorphism.CNV import CNVCompare, CNVSegmentBinarySearchTreeKey, get_overlap_ratio
		if tableObject is None:
			tableObject = self.getTableObject(tableName=tableName)
		sys.stderr.write("Constructing association-peak RBDict from HDF5 file %s, (peakPadding=%s) ..."%(self.inputFname, self.peakPadding))
		associationPeakRBDict = RBDict()
		associationPeakRBDict.result_id = None	#2012.6.22
		associationPeakRBDict.peakPadding = self.peakPadding
		associationPeakRBDict.HDF5AttributeNameLs = []
		
		for attributeName, value in self.getAttributes().items():
			associationPeakRBDict.HDF5AttributeNameLs.append(attributeName)
			setattr(associationPeakRBDict, attributeName, value)
		
		counter = 0
		real_counter = 0
		for row in tableObject:
			if not row['chromosome']:	#empty chromosome, which happens when inputFname contains no valid peaks, but the default null peak (only one).
				continue
			counter += 1
			segmentKey = CNVSegmentBinarySearchTreeKey(chromosome=row['chromosome'], \
							span_ls=[max(1, row['start'] - self.peakPadding), row['stop'] + self.peakPadding], \
							min_reciprocal_overlap=1, result_peak_id=None)
							#2010-8-17 overlapping keys are regarded as separate instances as long as they are not identical.
			if segmentKey not in associationPeakRBDict:
				associationPeakRBDict[segmentKey] = []
			else:
				sys.stderr.write("Warning: segmentKey of %s already in associationPeakRBDict with this row: %s.\n"%\
								(row, associationPeakRBDict[segmentKey][0]))
			associationPeakRBDict[segmentKey].append(castPyTablesRowIntoPassingData(row))	#row is a pointer to the current row.
		sys.stderr.write("%s peaks in %s spans.\n"%(counter, len(associationPeakRBDict)))
		
		self.associationPeakRBDict = associationPeakRBDict
		return self.associationPeakRBDict
	
	def appendAssociationPeak(self, association_peak_ls=None):
		"""
		2012.11.20
		"""
		sys.stderr.write("Dumping %s association peaks into %s ..."%(len(association_peak_ls), self.inputFname))
		#each number below is counting bytes, not bits
		cellList = []
		#2012.11.28 sort it
		association_peak_ls.sort()
		for association_peak in association_peak_ls:
			dataTuple = (association_peak.chromosome, association_peak.start, association_peak.stop, \
						association_peak.start_locus_id, association_peak.stop_locus_id, \
						association_peak.no_of_loci,\
						association_peak.peak_locus_id, association_peak.peak_score)
			self.associationPeakTable.writeOneCell(dataTuple)
		self.flush()
		sys.stderr.write(" \n")
	



"""
2013.2.12 an abstract class for reducer
"""

import sys, os, math
import copy
from palos import ProcessOptions, getListOutOfStr, PassingData, utils
from palos.mapper.AbstractMapper import AbstractMapper

class AbstractReducer(AbstractMapper):
    __doc__ = __doc__
    option_default_dict = copy.deepcopy(AbstractMapper.option_default_dict)
    option_default_dict.update({
        ('exitNonZeroIfAnyInputFileInexistent', 0, ): ['', '', 0, 
            "by default, it skips files that don't exist. Toggling this to exit 3.", ],\
        ('noHeader', 0, int): [0, 'n', 0, 
            'all input has no header'],\
        ('inputEmptyType', 0, int): [0, '', 1, 
           'when to call the input is empty.\n'
           '0: completely empty; 1: fastq empty, type 0 + a "@" only line '
           'followed by "+"-only line; 2: fasta empty, type 0+ a sequence ID line with only >.'],\
        })
    def __init__(self, inputFnameLs=None, **keywords):
        AbstractMapper.__init__(self, inputFnameLs=inputFnameLs, **keywords)
    
    def isInputLineEmpty(self, inputLine=None, inputFile=None, inputEmptyType=0):
        """
            only inputEmptyType 1 needs to access inputFile
        """
        if inputLine is None or inputLine=='':
            return True
        content = inputLine.strip()
        if content=='':
            return True
        
        answer = False
        if inputEmptyType==1:
            if content=='@':
                currentPosition = inputFile.tell()
                try:
                    nextLine = next(inputFile).strip()
                    afterNextLinePosition = inputFile.tell()
                    if nextLine.strip()=='+':
                        answer = True
                    else:
                        inputFile.seek(currentPosition-afterNextLinePosition, 1)
                        #rewind it back from the current position
                        """
                        use f.seek(offset, from_what). the reference point is selected by the from_what argument.
                        A from_what value of 0 measures from the beginning of the file, 1 uses the current file position,
                             and 2 uses the end of the file as the reference point.
                        """
                except StopIteration:
                    sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
                    import traceback
                    traceback.print_exc()
                    sys.stderr.write("End of file reached in trying to get '+'.\n")
                
        elif inputEmptyType==2:
            if content=='>':
                answer = True
        return answer


#!/usr/bin/env python3
"""
2007-10-01
	module to get the maximum clique
"""
import sys
sys.setrecursionlimit(50000)

class clique:
	def __init__(self, debug=0):
		self.debug = int(debug)
		self.max_clique_size = 0
		self.max_clique_ls = []
	
	def sub_max_clique(self, graph, node_ls, current_clique_size):
		"""
		2007-10-01
			algorithm 1 in Ostergard2002, which is Carraghan1990.
			get some clue about how to save the max clique from Ostergard's cliquer source code
		"""
		no_of_nodes = len(node_ls)
		if no_of_nodes==0:
			if current_clique_size>self.max_clique_size:
				self.max_clique_size = current_clique_size
				self.max_clique_ls = []
				return True
			return False
		return_value = False	#default is False
		while no_of_nodes!=0:
			if current_clique_size+no_of_nodes <= self.max_clique_size:
				return return_value
			v = node_ls.pop(0)	#smallest node
			new_node_ls = []
			for vj in node_ls:	#take the joint set between node_ls and neighbors of v
				if graph.has_edge(v, vj):
					new_node_ls.append(vj)
			if self.sub_max_clique(graph, new_node_ls, current_clique_size+1):
				self.max_clique_ls.append(v)
				if return_value==False:
					return_value = True	#as far as one of the max_clique() returns TRUE, this is successful. in order to trace up all nodes.
			no_of_nodes = len(node_ls)
		return return_value
	
	def order_nodes(self, graph):
		"""
		order the nodes according to degree
		"""
		node_ls = graph.nodes()
		node_degree_ls = graph.degree()
		import numpy
		argsort_node_degree_ls = numpy.argsort(node_degree_ls)
		new_node_ls = []
		for i in argsort_node_degree_ls:
			new_node_ls.append(node_ls[i])
		new_node_ls.reverse()
		return new_node_ls
	
	def init(self):
		self.max_clique_size = 0
		self.max_clique_ls = []
	
	def max_clique(self, graph):
		if self.debug:
			import pdb
			pdb.set_trace()
		self.init()
		node_ls = self.order_nodes(graph)
		self.sub_max_clique(graph, node_ls, 0)

if __name__ == '__main__':
	import networkx as nx
	G=nx.Graph()
	G.add_edge(1,2)
	G.add_edge(2,3)
	G.add_edge(2,4)
	G.add_edge(2,5)
	G.add_edge(2,6)
	G.add_edge(4,6)
	#G.add_edge(7,8)
	#G.add_edge(3,4)
	#G.add_edge(3,6)
	clique_instance = clique(debug=1)
	clique_instance.max_clique(G)
	print(f'max_clique_size: {clique_instance.max_clique_size}')
	print(f'max_clique_ls: {clique_instance.max_clique_ls}')
	print(f'nodes: {G.nodes()}')
	print(f'degree: {G.degree()}')
	nx.draw(G)
	import pylab
	print(f'the graph looks like this:')
	pylab.show()
	
	from networkx import cliques
	G_max_clique = cliques.find_cliques(G)
	print(f'cliques found  by networkx: {G_max_clique}')


#!/usr/bin/env python3
"""
2012.1.3
	module for graph algorithms/data structures
"""

import os, sys
sys.path.insert(0, os.path.expanduser('~/script'))
from palos.utils import PassingData
import copy
import numpy
from networkx import Graph, DiGraph
import networkx as nx

class GraphWrapper(Graph):
	def __init__(self, data=None, **keywords):
		"""
		2013.1.3
			wrapper around networkx.Graph
		"""
		Graph.__init__(self, data=None, **keywords)
		
	def recursiveTrimOutOfSetLeafNodes(self, nodeIdSet=None, recursiveDepth=0, total_no_of_nodes_removed=0):
		"""
		2013.1.2
			recursively remove all nodes with no out-degree.
			for directed graph. nodes with out_degree=0 are leaf nodes.
		"""
		if recursiveDepth==0:
			sys.stderr.write("Trimming leaf nodes that are not in given set (%s elements) ... \n"%\
						(len(nodeIdSet)))
		else:
			sys.stderr.write("%sdepth=%s"%('\x08'*80, recursiveDepth))
		no_of_nodes_removed = 0
		for n in self.nodes():
			if n not in nodeIdSet and self.out_degree(n)==0:
				self.remove_node(n)
				no_of_nodes_removed += 1
		total_no_of_nodes_removed += no_of_nodes_removed
		if no_of_nodes_removed>0:
			self.recursiveTrimOutOfSetLeafNodes(nodeIdSet=nodeIdSet, recursiveDepth=recursiveDepth+1,\
											total_no_of_nodes_removed=total_no_of_nodes_removed)
		else:
			sys.stderr.write("\n %s nodes removed."%total_no_of_nodes_removed)
		if recursiveDepth==0:
			sys.stderr.write("\n")
	
	def recursiveRemoveUniDegreeNodes(self):
		"""
		2013.1.2
			recursively remove all nodes whose degree (in + out) <= 1
		"""
		self.recursiveRemoveNodesByDegree(maxDegree=1, recursiveDepth=0)
	
	def recursiveRemoveNodesByDegree(self, maxDegree=0, recursiveDepth=0, total_no_of_nodes_removed=0, report=False):
		"""
		2013.1.2
			recursively remove all nodes whose degree (in + out) is below or equal to maxDegree 
		"""
		if report:
			if recursiveDepth==0:
				sys.stderr.write("Recursively remove nodes whose degree is <=%s ...\n"%maxDegree)
			else:
				sys.stderr.write("%sdepth=%s"%('\x08'*80, recursiveDepth))
		no_of_nodes_removed = 0
		for n in self.nodes():
			if self.degree(n)<=maxDegree:
				self.remove_node(n)
				no_of_nodes_removed += 1
		total_no_of_nodes_removed += no_of_nodes_removed
		if no_of_nodes_removed>0:
			self.recursiveRemoveNodesByDegree(maxDegree=maxDegree, recursiveDepth=recursiveDepth+1,\
											total_no_of_nodes_removed=total_no_of_nodes_removed)
		elif report:	#final round
				sys.stderr.write("\n %s nodes removed."%total_no_of_nodes_removed)
		if recursiveDepth==0 and report:
			sys.stderr.write("\n")
	
	def findAllLeafNodes(self):
		"""
		2013.1.3
		"""
		sys.stderr.write("Finding all leaf nodes ...")
		nodeList = []
		for n in self.nodes():
			if self.out_degree(n)==0:
				nodeList.append(n)
		sys.stderr.write(" %s leaves \n"%(len(nodeList)))
		return nodeList
	
	
	def _recursiveCalculateNodeDistanceToLeaf(self, leafNodes=None, node2distanceToLeaf=None):
		"""
		2013.10.16
		"""
		newLeafNodeSet = set()
		for leafNode in leafNodes:
			if leafNode not in node2distanceToLeaf:
				node2distanceToLeaf[leafNode] = 0
			for parent in self.predecessors(leafNode):
				node2distanceToLeaf[parent] = node2distanceToLeaf[leafNode] + 1
				newLeafNodeSet.add(parent)
		if newLeafNodeSet:
			self._recursiveCalculateNodeDistanceToLeaf(newLeafNodeSet, node2distanceToLeaf)
		return node2distanceToLeaf
	
	def calculateNodeHierarchyLevel(self):
		"""
		2013.1.3
			the hierarchy level for each node is the shortest path to all leaf nodes.
		"""
		
		leafNodes = set(self.findAllLeafNodes())
		sys.stderr.write("Calculating node hierarchy level ...")
		node2distanceToLeaf = {}
		#leaf nodes' hierarchy level=0
		self._recursiveCalculateNodeDistanceToLeaf(leafNodes, node2distanceToLeaf)
		"""
		for source in self.nodes():
			if source in leafNodes:
				node2distanceToLeaf[source] = 0
			else:
				level = None
				for target in leafNodes:
					try:
						l = nx.astar_path_length(self, source, target)
						if level is None:
							level = l
						elif l<level:
							level = l
					except:
						#no path between source and target
						pass
				node2distanceToLeaf[source] = level
		"""
		sys.stderr.write("%s nodes with hierarchy level.\n"%(len(node2distanceToLeaf)))
		return node2distanceToLeaf
	
	def findAllFounders(self):
		"""
		2013.3.5
			opposite of findAllLeafNodes()
		"""
		sys.stderr.write("Finding all founders ...")
		nodeList = []
		for n in self.nodes():
			if self.in_degree(n)==0:
				nodeList.append(n)
		sys.stderr.write(" %s founders \n"%(len(nodeList)))
		return nodeList
	
	def orderMembersByDistanceToFounders(self, founderSet=None):
		"""
		2013.3.5
			similar to calculateNodeHierarchyLevel
				but calculates shortest path to all founders (not leaf nodes)
					and the distance of the longest (not shortest) shortest path is the level.
				level of founder = 0
				level of founders' kids = 1
			This is good to simulate genotype of pedigree from founders to everyone.
		"""
		if founderSet is None:
			founderSet = set(self.findAllFounders())
		
		sys.stderr.write("Ordering members of the pedigree based on distance to founders ...")
		nodeID2distanceToFounder = {}
		#leaf nodes' hierarchy level=0
		counter = 0
		for source in self.nodes():
			if source in founderSet:
				nodeID2distanceToFounder[source] = 0
			else:
				level = None
				#find the longest path to the founder
				for founder in founderSet:
					try:
						l = nx.astar_path_length(self, founder, source)
						if level is None:
							level = l
						elif l>level:
							level = l
					except:
						#no path between source and founder
						pass
				nodeID2distanceToFounder[source] = level
			counter += 1
		sys.stderr.write("%s different hierarchy level.\n"%(len(nodeID2distanceToFounder)))
		return nodeID2distanceToFounder
	
	
class DiGraphWrapper(DiGraph, GraphWrapper):
	def __init__(self, data=None, **keywords):
		"""
		2013.3.5, put GraphWrapper, behind DiGraph in inheritance
		2013.1.3
		"""
		DiGraph.__init__(self, data=None, **keywords)
		self._undirectedG = None
	
	@property
	def undirectedGraph(self):
		"""
		2013.10.01
			assuming graph does not change anymore.
			
		"""
		if self._undirectedG is None:
			self._undirectedG = self.to_undirected()
		return self._undirectedG
	
	def getShortestPathInUndirectedVersion(self, nodeID1=None, nodeID2=None):
		"""
		2013.10.01
			get undirected version of graph, get shortest path between two nodes,
			then add edge direction to each edge
			
		"""
		shortestPath = nx.astar_path(self.undirectedGraph, source=nodeID1, target=nodeID2)
		pathWithDirection = []
		for i in range(len(shortestPath)-1):
			if self.has_edge(shortestPath[i], shortestPath[i+1]):
				edgeDirection = +1
			elif self.has_edge(shortestPath[i+1], shortestPath[i]):
				edgeDirection = -1
			else:
				sys.stderr.write("ERROR: Edge %s -> %s exists in undirected graph, but it (or its reverse) does not exist in directed version.\n"%
								(shortestPath[i], shortestPath[i+1]))
				raise
			pathWithDirection.append((shortestPath[i], shortestPath[i+1], edgeDirection))
		return pathWithDirection

if __name__ == '__main__':
	import pdb
	pdb.set_trace()
	



#!/usr/bin/env python3
"""
2012.6.5
    abstract class for db-interacting pegasus jobs
"""
import sys, os
import csv
from palos import ProcessOptions, getListOutOfStr, PassingData, utils
from palos.mapper import AbstractMapper
from . AbstractMapper import AbstractMapper

class AbstractDBJob(AbstractMapper):
    __doc__ = __doc__
    option_default_dict = AbstractMapper.option_default_dict.copy()
    #option_default_dict.pop(('inputFname', 1, ))
    option_default_dict.update({
        ('logFilename', 0, ): [None, '', 1, 'File to hold logs. '
            'Useful if a job is at the end of a pegasus workflow and '
            'has no other output file.'
            'Because otherwise, pegasus optimization will get rid of this job.'
            '(no output file, why need it?)'],\
        ('commit', 0, int):[0, '', 0, 'commit db transaction'],\
        ('sshTunnelCredential', 0, ): ['', '', 1, \
            'a ssh credential to allow machine to access db server. '
            'polyacti@login3, yuhuang@hpc-login2. '
            'If empty or port is empty, no tunnel',],
        })
    option_default_dict.update(AbstractMapper.db_option_dict)
    
    def __init__(self, inputFnameLs=None, **keywords):
        """
        2011-7-11
        """
        #self.connectDB() called within its __init__()
        AbstractMapper.__init__(self, inputFnameLs=inputFnameLs, **keywords)
        
        #2012.7.4 keep track of all the source&destination files,
        #  used by moveNewISQFileIntoDBStorage()
        self.srcFilenameLs = []
        self.dstFilenameLs = []
        #2012.11.18
        if getattr(self, "logFilename", None):
            self.logF = open(self.logFilename, 'w')
        else:
            self.logF = None
    
    def parseInputFile(self, inputFname=None, **keywords):
        """
        2013.08.23
            if a program is adding a file to db-affiliated storage,
             this is used for parsing.
        """
        return PassingData()
    
    def connectDB(self):
        """
        2012.5.11
            place holder.
        """
        pass
    
    def rmGivenFiles(self, filenameLs=[], rmCommand='rm -rf'):
        """
        2012.7.4
            delete all files in filenameLs
        """
        sys.stderr.write("Deleting %s files ...\n"%(len(filenameLs)))
        for filename in filenameLs:
            commandline = '%s %s'%(rmCommand, filename)
            return_data = utils.runLocalCommand(commandline, 
                report_stderr=True, report_stdout=True)
            if return_data.stderr_content:
                sys.stderr.write("commandline %s failed: %s\n"%(commandline,
                    return_data.stderr_content))
        sys.stderr.write(".\n")
    
    def cleanUpAndExitOnFailure(self, exitCode=1):
        """
        2012.7.13 an exit function when the program failed somewhere
        """
        #delete all target files.
        self.rmGivenFiles(filenameLs=self.dstFilenameLs)
        sys.exit(exitCode)
    
    def cleanUpAndExitOnSuccess(self, exitCode=0):
        """
        2012.7.13  an exit function when the program succeeded in the end
        """
        sys.exit(exitCode)
    
    def outputLogMessage(self, logMessage=None, logFilename=None, logF=None):
        """
        2012.11.18
            1. do not close _logF
            2. use self.logF if None available.
            3. append to logFilename if the latter is given.
        2012.7.17
        """
        _logF = None
        if logF is None:
            if logFilename:
                _logF = open(logFilename, 'a')
            elif self.logF:
                _logF = self.logF
        if _logF:
            _logF.write(logMessage)
    
    def closeLogF(self):
        """
        2012.11.18
        """
        if getattr(self, 'logF', None):
            self.logF.close()
    
    def sessionRollback(self, session=None):
        """
        2013.04.05
        wrap try...except around it because sometimes db connection is gone
            halfway through.
        and it causes program to exit without proper cleanup
        """
        try:
            session.rollback()
        except:
            sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
            import traceback
            traceback.print_exc()

    def run(self):
        pass


    


#!/usr/bin/env python3
"""
Examples:
	#setup database in mysql
	%s -v mysql -u yh -z papaya -d taxonomy -k ""
	
	# 2012.6.6 setup & import NCBI tax dump into db
	%s -u yh -k taxonomy -d vervetdb -v postgresql -i /usr/local/research_data/NCBI/taxonomy
	
Description:
	This is the taxonomy database ORM.
	The data is from taxdump in ftp://ftp.ncbi.nih.gov/pub/taxonomy/.
	It also imports the taxdump into db if given a folder of un-tarred dump files (-i).
"""
import sys, os
__doc__ = __doc__%(sys.argv[0], sys.argv[0])

from datetime import datetime
from sqlalchemy import DateTime, String, Integer
from sqlalchemy import Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column
from sqlalchemy import UniqueConstraint

from palos.db import Database, TableClass

Base = declarative_base()
#Set it staticaly because DB is undefined at this point 
# and it has to be defined after this.
_schemaname_ = "taxonomy"
class Citation(Base, TableClass):
	"""
	2012.6.6
	"""
	cit_id = Column(Integer)	#- the unique id of citation
	cit_key = Column(Text)	# -- citation key
	pubmed_id = Column(Integer)	#  -- unique id in PubMed database (0 if not in PubMed)
	medline_id = Column(Integer)		#  -- unique id in MedLine database (0 if not in MedLine)
	url = Column(Text)	# -- URL associated with citation
	text = Column(Text)	#-- any text (usually article name and authors).
		#-- The following characters are escaped in this text by a backslash:
		#-- newline (appear as "\n"),
		#-- tab character ("\t"),
		#-- double quotes ('\"'),
		#-- backslash character ("\\").
	tax_list = ManyToMany("Node", tablename='citation2node', local_colname='citation_id')
		# -- list of node ids separated by a single space
	created_by = Column(String(256))
	updated_by = Column(String(256))
	date_created = Column(DateTime, default=datetime.now)
	date_updated = Column(DateTime)
	using_options(tablename='citation')
	using_table_options(mysql_engine='InnoDB')

class Delnode(Base, TableClass):
	"""
	2012.6.6
	"""
	tax_id = Column(Integer, unique=True)	# -- taxonomy database division id
	created_by = Column(String(256))
	updated_by = Column(String(256))
	date_created = Column(DateTime, default=datetime.now)
	date_updated = Column(DateTime)
	using_options(tablename='delnode')
	using_table_options(mysql_engine='InnoDB')

class Division(Base, TableClass):
	"""
	2012.6.6
	"""
	id = Column(Integer, primary_key=True)	# -- taxonomy database division id
	code = Column(Text)	# -- GenBank division code (three characters)
		# e.g. BCT, PLN, VRT, MAM, PRI...
	name = Column(Text)
	comments = Column(Text)
	created_by = Column(String(256))
	updated_by = Column(String(256))
	date_created = Column(DateTime, default=datetime.now)
	date_updated = Column(DateTime)
	using_options(tablename='division')
	using_table_options(mysql_engine='InnoDB')

class Gencode(Base, TableClass):
	"""
	"""
	id = Column(Integer, primary_key=True)	# -- GenBank genetic code id
	abbreviation = Column(Text)	# -- genetic code name abbreviation
	name = Column(Text)	# -- genetic code name
	code = Column(Text)	#-- translation table for this genetic code
	starts = Column(Text)	#	-- start codons for this genetic code
	comment = Column(Text)
	created_by = Column(String(256))
	updated_by = Column(String(256))
	date_created = Column(DateTime, default=datetime.now)
	date_updated = Column(DateTime)
	using_options(tablename='gencode')
	using_table_options(mysql_engine='InnoDB')

class Merged(Base, TableClass):
	"""
	2012.6.6
	"""
	old_tax_id = Column(Integer)
	#-- id of nodes which has been merged
	new_tax = ManyToOne('%s.Node'%__name__, colname='new_tax_id', ondelete='CASCADE', onupdate='CASCADE')
		#-- id of nodes which is result of merging
	created_by = Column(String(256))
	updated_by = Column(String(256))
	date_created = Column(DateTime, default=datetime.now)
	date_updated = Column(DateTime)
	using_options(tablename='merged')
	using_table_options(mysql_engine='InnoDB')


class Name(Base, TableClass):
	"""
	2012.6.6
		taxonomy name
	"""
	tax = ManyToOne('%s.Node'%__name__, colname='tax_id', ondelete='CASCADE', onupdate='CASCADE')
	name_txt = Column(Text)
	unique_name = Column(Text)
	name_class = Column(Text)
	created_by = Column(String(256))
	updated_by = Column(String(256))
	date_created = Column(DateTime, default=datetime.now)
	date_updated = Column(DateTime)
	using_options(tablename='name')
	using_table_options(mysql_engine='InnoDB')
	using_table_options(UniqueConstraint('tax_id','name_txt','name_class'))
	

class Node(Base, TableClass):
	"""
	2012.6.6
		taxonomy node and its parent node
	"""
	tax_id = Column(Integer, unique=True, primary_key=True)	#-- node id in GenBank taxonomy database
	parent_tax_id = Column(Integer)
	#2012.6.7 some parent taxonomy node might not exist at all.
	#parent_tax = ManyToOne('%s.Node'%__name__, colname='parent_tax_id', ondelete='CASCADE', onupdate='CASCADE')
	rank = Column(Text)	#-- rank of this node (superkingdom, kingdom, ...)
	embl_code = Column(Text)	#-- locus-name prefix; not unique
	division = ManyToOne('%s.Division'%__name__, colname='division_id', ondelete='CASCADE', onupdate='CASCADE')
		#see division.dmp file
	inherited_div_flag = Column(Integer)
	#  (1 or 0) -- 1 if node inherits division from parent
	genetic_code = ManyToOne('%s.Gencode'%__name__, colname='genetic_code_id', ondelete='CASCADE', onupdate='CASCADE')
		# -- see gencode.dmp file
	inherited_GC_flag = Column(Integer)
	#(1 or 0) -- 1 if node inherits genetic code from parent
	mitochondrial_genetic_code = ManyToOne('%s.Gencode'%__name__, colname='mitochondrial_genetic_code_id', \
										ondelete='CASCADE', onupdate='CASCADE')
		# -- see gencode.dmp file
	inherited_MGC_flag = Column(Integer)
	#(1 or 0)	#  -- 1 if node inherits mitochondrial gencode from parent
	GenBank_hidden_flag = Column(Integer)
	# (1 or 0) -- 1 if name is suppressed in GenBank entry lineage
	hidden_subtree_root_flag = Column(Integer)	
	#(1 or 0) -- 1 if this subtree has no sequence data yet
	comments = Column(Text)
	citation_list = ManyToMany("Citation", tablename='citation2node', local_colname='tax_id')
	created_by = Column(String(256))
	updated_by = Column(String(256))
	date_created = Column(DateTime, default=datetime.now)
	date_updated = Column(DateTime)
	using_options(tablename='node')
	using_table_options(mysql_engine='InnoDB')

class TaxonomyDB(Database):
	__doc__ = __doc__
	option_default_dict = Database.option_default_dict.copy()
	option_default_dict[('drivername', 1,)][0] = 'postgresql'
	option_default_dict[('dbname', 1,)][0] = 'taxonomy'
	option_default_dict.update({
		('inputFolder', 0, ):[None, 'i', 1, 
		'where taxdump.tar.gz was un-tarred. it should contain citations.dmp, division.dmp, '
		'gencode.dmp, names.dmp, delnodes.dmp, merged.dmp, nodes.dmp.'
		'If this argument is provided, this program will import them into a taxonomy database.\n'],
		('commit', 0, int):[0, 'c', 0, 'commit db transaction'],\
		})	
	def __init__(self, **keywords):
		"""
		2008-10-08
			simplified further by moving db-common lines to Database
		2008-07-09
		"""
		Database.__init__(self, **keywords)
		self.setup_engine(metadata=__metadata__, session=__session__, entities=entities)
		self.ncbiTaxDumpFileDelimiter = '\t|\t'
		
		self._scientific_name2tax_id = None
		self._tax_id2scientific_name = None
	
	def splitLine(self, line=None):
		line = line.strip()
		if line[-2:]=='\t|':	#the last field has this ending
			line = line[:-2]
		row = line.split(self.ncbiTaxDumpFileDelimiter)
		return row
	
	def importMergedTaxNodes(self, inputFile=None):
		"""
		2012.6.6
		"""
		sys.stderr.write('Importing merged taxonomy nodes ...')
		counter = 0
		for line in inputFile:
			row = self.splitLine(line)
			old_tax_id = int(row[0])
			new_tax_id = int(row[1])
			db_entry = Merged(old_tax_id=old_tax_id, new_tax_id=new_tax_id)
			self.session.add(db_entry)
			counter += 1
		self.session.flush()
		sys.stderr.write("%s merged nodes added into db.\n"%(counter))
	
	def importTaxNodes(self, inputFile=None, divisionID2division=None, gencodeID2gencode=None):
		"""
		2012.6.6
			create a structure of parentTaxonomy2Child first to insure parent taxonomy is in db first
		"""
		sys.stderr.write('Importing taxonomy nodes ...\n')
		counter = 0
		import networkx as nx
		DG = nx.DiGraph()	#use this to store the relationship among them
		DG.add_weighted_edges_from([(1,2,0.5), (3,1,0.75)])
		row_ls = []
		taxID2rowIndex = {}
		for line in inputFile:
			row = self.splitLine(line)
			
			tax_id = int(row[0])
			parent_tax_id = int(row[1])
			taxID2rowIndex[tax_id] = len(row_ls)
			row_ls.append(row)
			if parent_tax_id==tax_id:	#the first taxonomy node has itself as the parent, which makes the graph not a DAG.
				sys.stderr.write("Warning: taxonomy ID %s's parent is itself. Not added in the directed graph.\n"%(tax_id))
			else:
				DG.add_edge(parent_tax_id, tax_id)
		G = DG.to_undirected()
		sys.stderr.write(" %s nodes, %s edges, %s connected components.\n"%(DG.number_of_nodes(), DG.number_of_edges(),\
			nx.number_connected_components(G)))

		#first check if this is DAG , it should be.
		if not nx.is_directed_acyclic_graph(DG):
			sys.stderr.write("Error: tax node graph is not a directed acyclic graph. it should be.\n")
			#sys.exit(3)
		# then save the nodes in the topological sort order (parents get saved before children).
		taxIDInTopoSortOrder = nx.topological_sort(DG)
		for tax_id in taxIDInTopoSortOrder:
			if tax_id not in taxID2rowIndex:
				sys.stderr.write("Warning: taxonomy ID %s has no real data in the input file.\n"%(tax_id))
				continue
			rowIndex = taxID2rowIndex.get(tax_id)
			row = row_ls[rowIndex]
			tax_id = int(row[0])
			parent_tax_id = int(row[1])
			rank = row[2]
			embl_code = row[3]
			division_id = int(row[4])
			
			division = divisionID2division[division_id]
			
			inherited_div_flag = int(row[5])
			genetic_code_id = int(row[6])
			genetic_code = gencodeID2gencode[genetic_code_id]
			
			inherited_GC_flag = int(row[7])
			mitochondrial_genetic_code_id = int(row[8])
			mitochondrial_genetic_code = gencodeID2gencode[mitochondrial_genetic_code_id]
			
			inherited_MGC_flag = int(row[9])
			GenBank_hidden_flag = int(row[10])
			hidden_subtree_root_flag = int(row[11])
			comments = row[12].decode('latin1')
			
			node = Node(tax_id=tax_id, parent_tax_id=parent_tax_id, rank=rank,
				embl_code=embl_code, inherited_div_flag=inherited_div_flag,\
				inherited_GC_flag=inherited_GC_flag,
				inherited_MGC_flag=inherited_MGC_flag,
				GenBank_hidden_flag=GenBank_hidden_flag,
				hidden_subtree_root_flag=hidden_subtree_root_flag,
				comments=comments)
			node.division = division
			node.genetic_code = genetic_code
			node.mitochondrial_genetic_code = mitochondrial_genetic_code
			self.session.add(node)
			counter += 1
			if counter%5000==0:
				sys.stderr.write("%s%s"%('\x08'*40, counter))
				self.session.flush()
		sys.stderr.write("%s%s\n"%('\x08'*40, counter))
		self.session.flush()
		sys.stderr.write("%s nodes added into db.\n"%(counter))
	
	def importDeletedNodes(self, inputFile=None):
		"""
		2012.6.6
		"""
		sys.stderr.write('Importing deleted nodes ...\n')
		counter = 0
		for line in inputFile:
			row = self.splitLine(line)
			tax_id = int(row[0])
			db_entry = Delnode(tax_id=tax_id)
			self.session.add(db_entry)
			self.session.flush()
			counter += 1
			if counter%5000==0:
				sys.stderr.write("%s%s"%('\x08'*40, counter))
		sys.stderr.write("%s%s\n"%('\x08'*40, counter))
		sys.stderr.write("%s added into db.\n"%(counter))
	
	def importTaxNames(self, inputFile=None):
		"""
		2012.6.6
		"""
		sys.stderr.write('Importing taxonomy names ...\n')
		counter = 0
		for line in inputFile:
			row = self.splitLine(line)
			tax_id = int(row[0])
			name_txt = row[1]
			unique_name = row[2]
			name_class = row[3]
			db_entry = Name(tax_id=tax_id, name_txt=name_txt,
				unique_name=unique_name, name_class=name_class)
			self.session.add(db_entry)
			counter += 1
			if counter%5000==0:
				sys.stderr.write("%s%s"%('\x08'*40, counter))
				self.session.flush()
		sys.stderr.write("%s%s\n"%('\x08'*40, counter))
		self.session.flush()
		sys.stderr.write("%s added into db.\n"%(counter))
	
	def importGeneticCode(self, inputFile=None):
		"""
		2012.6.6
		"""
		sys.stderr.write('Importing genetic codes ...')
		counter = 0
		gencodeID2gencode = {}
		for line in inputFile:
			row = self.splitLine(line)
			gencode_id = int(row[0])
			abbr = row[1]
			name = row[2]
			code = row[3]
			starts = row[4]
			gencode = Gencode(id=gencode_id, abbreviation=abbr, name=name,
				code=code, starts=starts)
			self.session.add(gencode)
			self.session.flush()
			gencodeID2gencode[gencode.id] = gencode
			counter += 1
		sys.stderr.write("%s added into db.\n"%(counter))
		return gencodeID2gencode
	
	
	def importDivision(self, inputFile=None):
		"""
		2012.6.6
		"""
		sys.stderr.write('Importing divisions ...')
		counter = 0
		divisionID2division = {}
		for line in inputFile:
			row = self.splitLine(line)
			division = Division(id=int(row[0]), code=row[1], name=row[2],
				comments=row[3])
			self.session.add(division)
			self.session.flush()
			divisionID2division[division.id] = division
			counter += 1
		sys.stderr.write("%s added into db.\n"%(counter))
		return divisionID2division
	
	def importCitation(self, inputFile=None):
		"""
		2012.6.6
		"""
		sys.stderr.write('Importing citations ...\n')
		counter = 0
		for line in inputFile:
			row = self.splitLine(line)
			cit_id = int(row[0])
			cit_key = row[1].decode('latin1')
			pubmed_id = int(row[2])
			medline_id = int(row[3])
			url = row[4].decode('latin1')
			if len(row)>5:
				text = row[5].decode('latin1')
				#the input file contains utf-8 characters beyond 127
				#  ('ascii' codec limit). 
				#.encode('ascii','replace')
			else:
				text = None
			if len(row)>6:
				taxIDList = row[6].split()
				taxIDList = map(int, taxIDList)
			else:
				taxIDList = []
			db_entry = Citation(cit_id=cit_id, cit_key=cit_key,
				pubmed_id=pubmed_id, medline_id=medline_id, url=url,
				text=text)
			for taxID in taxIDList:
				tax = Node.get(taxID)
				if tax is None:
					sys.stderr.write("Error: taxonomy ID %s not in table node.\n"%(taxID))
					sys.exit(2)
				db_entry.tax_list.append(tax)
			self.session.add(db_entry)
			counter += 1
			if counter%5000==0:
				sys.stderr.write("%s%s"%('\x08'*40, counter))
				self.session.flush()
		sys.stderr.write("%s%s\n"%('\x08'*40, counter))
		self.session.flush()
		sys.stderr.write("%s added into db.\n"%(counter))
	
	
	def importNCBITaxDump(self, inputFolder=None, ):
		"""
		2012.6.6
			
		"""
		sys.stderr.write("Importing NCBI tax dump files from %s ...\n"%\
			(inputFolder))
		citationsFile = open(os.path.join(inputFolder, 'citations.dmp'), 'r')
		divisionFile = open(os.path.join(inputFolder, 'division.dmp'), 'r')
		gencodeFile = open(os.path.join(inputFolder, 'gencode.dmp'), 'r')
		namesFile = open(os.path.join(inputFolder, 'names.dmp'), 'r')
		delnodesFile = open(os.path.join(inputFolder, 'delnodes.dmp'), 'r')
		mergedFile = open(os.path.join(inputFolder, 'merged.dmp'), 'r')
		nodesFile = open(os.path.join(inputFolder, 'nodes.dmp'), 'r')
		openedFileList = [citationsFile, divisionFile, gencodeFile, namesFile,
			delnodesFile, mergedFile, nodesFile]
		
		gencodeID2gencode = self.importGeneticCode(gencodeFile)
		divisionID2division = self.importDivision(divisionFile)
		self.importTaxNodes(nodesFile, gencodeID2gencode=gencodeID2gencode,
			divisionID2division=divisionID2division)
		
		self.importTaxNames(namesFile)
		self.importMergedTaxNodes(mergedFile)
		self.importDeletedNodes(delnodesFile)
		self.importCitation(citationsFile)
		
		for openedFile in openedFileList:
			openedFile.close()
		
		sys.stderr.write("\n")
	
	def run(self):
		"""
		"""
		if self.debug:
			import pdb
			pdb.set_trace()
	
		if self.inputFolder:	#only pick the file if the output file is not empty.
			self.session.begin()
			self.importNCBITaxDump(self.inputFolder)
			if self.commit:
				self.session.flush()
				self.session.commit()
			else:
				self.session.rollback()
	
	@property
	def scientific_name2tax_id(self):
		"""
		2012.8.28
			copied from palos/utils.py
		2012.6.6
			update it to get table names from TaxonomyDB
		"""
		if self._scientific_name2tax_id is None:
			self.fillUpMapBetweenTaxIDAndScientificName()
		return self._scientific_name2tax_id
	
	def returnTaxIDGivenScientificName(self, scientific_name=None):
		"""
		2012.8.28
			copied from palos/utils.py
		"""
		return self.scientific_name2tax_id.get(scientific_name)
	
	def returnTaxIDGivenSentence(self, sentence=None):
		"""
		2012.8.28
			copied from palos/utils.py
		2008-07-29
		"""
		tax_id_to_return = None
		for scientific_name, tax_id in self.scientific_name2tax_id.items():
			if sentence.find(scientific_name)>=0:
				tax_id_to_return = tax_id
				break
		return tax_id_to_return
	
	def fillUpMapBetweenTaxIDAndScientificName(self):
		"""
		2012.8.29
			remove " and o.rank='species'" from the query condition
		2012.8.29
		"""
		self._scientific_name2tax_id = {}
		self._tax_id2scientific_name = {}
		curs = self.metadata.bind
		rows = curs.execute("SELECT n.name_txt, n.tax_id FROM %s.%s n, %s.%s o where n.name_class='scientific name' \
			and n.tax_id=o.tax_id"%(self.schema, Name.table.name, \
			self.schema, Node.table.name))
		#rows = curs.fetchall()
		for row in rows:
			scientific_name = row.name_txt
			tax_id = row.tax_id
			self._tax_id2scientific_name[tax_id] = scientific_name
			self._scientific_name2tax_id[scientific_name] = tax_id
		sys.stderr.write("%s entries in _tax_id2scientific_name. %s entries in _scientific_name2tax_id.\n"%\
			(len(self._tax_id2scientific_name), len(self._scientific_name2tax_id)))

	@property
	def tax_id2scientific_name(self):
		"""
		2012.8.28
		"""
		if self._tax_id2scientific_name is None:
			self.fillUpMapBetweenTaxIDAndScientificName()
		return self._tax_id2scientific_name
	
	def returnScientificNameGivenTaxID(self, tax_id=None):
		"""
		2012.8.28
			there's a cache involved in the 1st query.
		"""
		tax_id2scientific_name = self.tax_id2scientific_name
		return tax_id2scientific_name.get(tax_id)
	
	def getScientificNameGivenTaxID(self, tax_id=None):
		"""
		2012.8.28
			no cache, jsut db query
		"""
		curs = self.metadata.bind
		rows = curs.execute("SELECT n.name_txt, n.tax_id FROM %s.%s n, %s.%s o where n.name_class='scientific name' \
				and n.tax_id=o.tax_id and n.tax_id=%s"%(self.schema, Name.table.name, \
				self.schema, Node.table.name, tax_id))
		row = rows.fetchone()
		if row:
			scientific_name = row.name_txt
			return scientific_name
		else:
			return None
	
	def getTaxIDGivenScientificName(self, scientific_name=None):
		"""
		2012.8.29
			no cache, just db query
		"""
		curs = self.metadata.bind
		rows = curs.execute("SELECT n.name_txt, n.tax_id FROM %s.%s n, %s.%s o where n.name_class='scientific name' \
				and n.tax_id=o.tax_id and n.name_txt='%s'"%(self.schema, Name.table.name,\
				self.schema, Node.table.name, scientific_name))
		row = rows.fetchone()
		if row:
			return row.tax_id
		else:
			return None
	
if __name__ == '__main__':
	import sys, os
	from palos import ProcessOptions
	main_class = TaxonomyDB
	po = ProcessOptions(sys.argv, main_class.option_default_dict, error_doc=main_class.__doc__)
	
	instance = main_class(**po.long_option2value)
	instance.setup()
	instance.run()



#!/usr/bin/env python3
"""
2008-10-07 simple functions related to matplotlib
"""
import os,sys
from scipy import stats

def assignMatPlotlibHueColorToLs(name_ls, debug=0):
	"""
	2008-10-07
		assign continuous HSL color spectrum to a list (in order).
		
	"""
	if debug:
		sys.stderr.write("Assigning matplotlib hue color to a list ...")
	import ImageColor
	no_of_names = len(name_ls)
	value_step = 1./(no_of_names-1)	#-1 to cover both minimum and maximum in the spectrum
	name2fc = {}
	for i in range(no_of_names):
		name = name_ls[i]
		value = i*value_step
		hue_value = max(min(int(round((1-value)*255)), 255), 0)	#max(min()) makes sure it's 0-255
		fc = ImageColor.getrgb('hsl(%s'%hue_value+',100%,50%)')#"hsl(hue, saturation%, lightness%)" where hue is the colour given as an
		# angle between 0 and 360 (red=0, green=120, blue=240),
		#saturation is a value between 0% and 100% (gray=0%, full color=100%), and lightness is a value between 0% and 100% (black=0%, normal=50%, white=100%).
		fc = [color_value/255. for color_value in fc]	#matplotlib accepts rgb in [0-1] range
		name2fc[name] = fc
	if debug:
		sys.stderr.write("Done.\n")
	return name2fc

def drawName2FCLegend(ax, name_ls, name2fc=None, shape_type=1, no_face_color=False, no_edge_color=False, title=None, font_size=4, alpha=1):
	"""
	2008-10-08
		add option title and linewidth
	2008-10-07
		draw a legend according to name_ls and colors according to name2fc.
		the shape of the legend is symmetric. It's either circle or square.
	"""
	sys.stderr.write("Drawing name2fc legend  ...")
	import matplotlib
	from matplotlib.patches import Polygon, Circle, Ellipse, Wedge
	import numpy
	
	if name2fc is None:
		name2fc = assignMatPlotlibHueColorToLs(name_ls)
	no_of_names = len(name_ls)
	value_step = min(1./(no_of_names), 0.5)	#can't exceed half of plot
	
	radius = value_step/4.
	center_x_pos = 0.25
	center_y_pos = value_step/2.
	xs = [center_x_pos-radius, center_x_pos+radius, center_x_pos+radius, center_x_pos-radius]	#X-axis value for the 4 points of the rectangle starting from lower left corner. 
	ys = numpy.array([center_y_pos-radius, center_y_pos-radius, center_y_pos+radius, center_y_pos+radius])
	for i in range(no_of_names):
		name = name_ls[i]
		fc = name2fc[name]
		facecolor = fc
		edgecolor = fc
		
		if no_face_color:
			facecolor='w'
		if no_edge_color:
			edgecolor='w'
			linewidth = 0
		else:
			linewidth = matplotlib.rcParams['patch.linewidth']
		if shape_type==1:
			patch = Circle((center_x_pos, center_y_pos), radius=radius, linewidth=linewidth, facecolor=facecolor, edgecolor=edgecolor, alpha=alpha)
			center_y_pos += value_step
		elif shape_type==2:
			patch = Polygon(zip(xs, ys), linewidth=linewidth, facecolor=facecolor, edgecolor=edgecolor, alpha=alpha)
			ys += value_step	#increase y-axis
		else:
			patch = Circle((center_x_pos, center_y_pos), radius=radius, linewidth=linewidth, facecolor=facecolor, edgecolor=edgecolor, alpha=alpha)
			center_y_pos += value_step
		ax.add_patch(patch)
		ax.text(0.75, center_y_pos-value_step, name, horizontalalignment ='left', verticalalignment='center', size=font_size)
	if title:
		ax.set_title(title, fontsize=font_size)
	sys.stderr.write("Done.\n")

def autoLabelBarChartWithHeight(axe, rects):
	"""
	2010-4-1
		from http://matplotlib.sourceforge.net/examples/pylab_examples/barchart_demo.html
		
		attach some text labels for each rectangular in rects (return of pylab.bar())
	"""
	for rect in rects:
		height = rect.get_height()
		axe.text(rect.get_x()+rect.get_width()/2., 1.05*height, '%.2f'%float(height), ha='center', va='bottom', fontsize='xx-small')

def setFontAndLabelSize(base_size=9):
	"""
	2012.12.3 make the legend & axes.title font size 1.4 X base_size
	2010-4-14
		function to change matplotlib font size globally
		#have to be set in the very beginning of a program, otherwise, no effect.
		
		for more additional customization check: http://matplotlib.sourceforge.net/users/customizing.html
	"""
	from matplotlib import rcParams
	rcParams['font.size'] = base_size	#default is 12
	rcParams['legend.fontsize'] = int(base_size*1.6)	#default is large 
	#rcParams['text.fontsize'] = 6	#deprecated. use font.size instead
	rcParams['axes.labelsize'] = base_size	#default is medium
	rcParams['axes.titlesize'] = int(base_size*1.6)	#default is large 
	rcParams['xtick.labelsize'] = base_size	#default is medium
	rcParams['ytick.labelsize'] = base_size	#default is medium

def setDefaultFigureSize(figsize=None):
	"""
	2012.9.6
		function to change matplotlib default figure size globally
		#have to be set in the very beginning of a program, otherwise, no effect.
		for more additional customization check: http://matplotlib.sourceforge.net/users/customizing.html
		
		below is a different way to changing figure size. not sure how inch translates into pixels.
			fig = matplotlib.pyplot.gcf()
			fig.set_size_inches(185, 60)
		
		For this figsize (X,Y) if dpi=N, then final dimension is (X*N, Y*N)
	"""
	from matplotlib import rcParams
	if figsize is None:
		figsize = (10,10)
	rcParams['figure.figsize'] = figsize

def setPlotDimension(left=0.1, right=0.9, bottom=0.1, top=0.9):
	"""
	2012.12.4
	
	for more additional customization check: http://matplotlib.sourceforge.net/users/customizing.html
		
	# The figure subplot parameters.  All dimensions are a fraction of the
	# figure width or height
	#figure.subplot.left    : 0.125  # the left side of the subplots of the figure
	#figure.subplot.right   : 0.9    # the right side of the subplots of the figure
	#figure.subplot.bottom  : 0.1    # the bottom of the subplots of the figure
	#figure.subplot.top     : 0.9    # the top of the subplots of the figure
	#figure.subplot.wspace  : 0.2    # the amount of width reserved for blank space between subplots
	#figure.subplot.hspace  : 0.2    # the amount of height reserved for white space between subplots
	
	could also be in set as 
		axe_pvalue = pylab.axes([-0.1, -0.1, 1.2, 1.2], frameon=False)	#left gap, bottom gap, width, height.
		pylab.figure(axe_pvalue.figure.number)
	"""
	from matplotlib import rcParams
	rcParams['figure.subplot.left'] = left
	rcParams['figure.subplot.right'] = right
	rcParams['figure.subplot.bottom'] = bottom
	rcParams['figure.subplot.top'] = top
	


def restoreMatplotlibRCDefaults():
	"""
	2010-4-14
		counter part of setFontAndLabelSize()
	"""
	import matplotlib
	matplotlib.rcdefaults()

def drawHist(data_ls=None, title=None, xlabel_1D=None, xticks=None, outputFname=None, min_no_of_data_points=50, needLog=False, \
			dpi=200, max_no_of_bins=20, **kwargs):
	"""
	2012.11.27 rename max_no_of_data_points to min_no_of_data_points
	2012.8.6 argument min_no_of_bins renamed to max_no_of_bins, which is actually what it does.
	2011-11-28
		add argument min_no_of_bins
	2011-8-24
		add argument kwargs, xticks
	2011-4-18
		a wrapper for histogram drawing using matplotlib
	"""
	sys.stderr.write("Drawing histogram of %s data points to %s..."%(len(data_ls), outputFname))
	import pylab
	pylab.clf()
	fig = pylab.figure(figsize=(10, 10))
	#ax = pylab.axes()
	#ax = fig.gca()
	no_of_data_points = len(data_ls)
	if no_of_data_points>=min_no_of_data_points:
		no_of_bins = max(10, min(max_no_of_bins, no_of_data_points/10))
		n, bins, patches = pylab.hist(data_ls, no_of_bins, log=needLog)
		if title:
			pylab.title(title)
		else:
			title = constructTitleFromDataSummaryStat(data_ls)
			#title = 'n=%s, mean %.4f, median %.4f'%(len(data_ls), meanValue, medianValue)
			pylab.title(title)
		if xlabel_1D is not None:
			pylab.xlabel(xlabel_1D)
		if xticks:
			x_ls = bins[:-1]	#the bins has one extra element.
			pylab.xticks(x_ls, xticks)
		pylab.savefig(outputFname, dpi=dpi)
	
	sys.stderr.write("Done.\n")

def drawBarChart(x_ls, y_ls, title=None, xlabel_1D=None, xticks=None, outputFname=None, bottom=0, needLog=False, dpi=200, **kwargs):
	"""
	2011-8-15
		a wrapper for barChart drawing using matplotlib
	"""
	sys.stderr.write("Drawing barChart of %s data points to %s..."%(len(y_ls), outputFname))
	import pylab
	pylab.clf()
	pylab.bar(x_ls, y_ls, width=0.8, bottom=bottom, log=needLog, **kwargs)
	pylab.title(title)
	if xlabel_1D is not None:
		pylab.xlabel(xlabel_1D)
	if xticks:
		pylab.xticks(x_ls, xticks)
	pylab.savefig(outputFname, dpi=dpi)
	
	sys.stderr.write("Done.\n")

def drawBoxPlot(data_2D_ls, title=None, xlabel_1D=None, xticks=None, outputFname=None, dpi=200, **kwargs):
	"""
	2011-8-15
		a wrapper for boxplot drawing using matplotlib
		
		data_2D_ls could be a list of lists or a 2D array.
			The former is more efficient because boxplot converts
			a 2-D array into a list of vectors internally anyway.
	"""
	sys.stderr.write("Drawing boxplots of %s data points to %s..."%(len(data_2D_ls), outputFname))
	import pylab
	pylab.clf()
	pylab.boxplot(data_2D_ls, notch=0, sym='+', vert=1, whis=1.5, positions=None, \
				widths=None, **kwargs)
	#patch_artist = False (default) produces boxes with the Line2D artist
	#patch_artist = True produces boxes with the Patch artist
	pylab.title(title)
	if xlabel_1D is not None:
		pylab.xlabel(xlabel_1D)
	if xticks:
		pylab.xticks(range(len(xticks)), xticks)
	pylab.savefig(outputFname, dpi=dpi)
	
	sys.stderr.write("Done.\n")

def logSum(ls):
	"""
	2011-4-27
		given a list, take the sum and then log10(sum).
		an alternative to the reduce_C_function argument in drawHexbin()
	"""
	import math
	return math.log10(sum(ls))

def drawHexbin(x_ls, y_ls, C_ls, fig_fname=None, gridsize=100, title=None, xlabel=None, ylabel=None,\
			colorBarLabel=None, reduce_C_function=None, dpi=300, mincnt=None, marginals=False, xscale='linear',\
			scale='linear'):
	"""
	2011.12.22
		xscale: [ linear | log]
			Use a linear or log10 scale on the horizontal axis.
		scale: [ linear | log]
			Use a linear or log10 scale on the vertical axis.
		mincnt: None | a positive integer
			If not None, only display cells with more than mincnt number of points in the cell
		marginals: True|False
			if marginals is True, plot the marginal density as colormapped rectagles along the bottom of the x-axis and left of the y-axis
	2011-4-27
		draw 2D histogram (reduce_C_function=logSum) or any 3D plot (3rd Dimension is determined by reduce_C_function).
		default of reduce_C_function: numpy.median.
		moved from variation/src/misc.py
	2010-7-1
		add argument reduce_C_function()
	2010-6-28
		add argument xlabel & ylabel
	2010-5-11
	"""
	import pylab, numpy
	import matplotlib.cm as cm
	if reduce_C_function is None:
		reduce_C_function = numpy.median
	pylab.clf()
	pylab.hexbin(x_ls, y_ls, C=C_ls, gridsize=gridsize, reduce_C_function=reduce_C_function, cmap=cm.jet,\
				mincnt=mincnt, marginals=marginals, xscale=xscale)#, scale=scale
	pylab.axis([min(x_ls), max(x_ls), min(y_ls), max(y_ls)])
	if title is None:
		title = "gridsize %s, %s probes."%(gridsize, len(x_ls))
	pylab.title(title)
	if xlabel:
		pylab.xlabel(xlabel)
	if ylabel:
		pylab.ylabel(ylabel)
	cb = pylab.colorbar()
	if colorBarLabel:
		cb.set_label(colorBarLabel)
	if fig_fname:
		pylab.savefig(fig_fname, dpi=dpi)



def drawScatter(x_ls=None, y_ls=None, fig_fname=None, title=None, xlabel=None, ylabel=None,\
			dpi=300, logX=False, logY=False, **keywords):
	"""
	2013.07.12 added argument logX, logY
	2011-10-17
	"""
	sys.stderr.write("Drawing scatter of %s points vs %s points to %s ..."%(len(x_ls), len(y_ls), fig_fname))
	import pylab, numpy
	import matplotlib.cm as cm
	pylab.clf()
	fig = pylab.figure(figsize=(10, 10)) # You were missing the =
	ax = fig.add_subplot(1, 1, 1)
	if logY:
		ax.set_yscale('log')
	if logX:
		ax.set_xscale('log')
	pylab.plot(x_ls, y_ls, '.', **keywords)
	if title is None:
		title = constructTitleFromTwoDataSummaryStat(x_ls, y_ls)
	pylab.title(title)
	if xlabel:
		pylab.xlabel(xlabel)
	if ylabel:
		pylab.ylabel(ylabel)
	if fig_fname:
		pylab.savefig(fig_fname, dpi=dpi)
	sys.stderr.write("Done.\n")

def constructTitleFromTwoDataSummaryStat(x_ls=None, y_ls=None):
	"""
	2015.01.28 added wilcoxon signed rank test
	2012.8.21
	
	"""
	if not y_ls and x_ls:
		title = constructTitleFromDataSummaryStat(x_ls)
	elif x_ls and y_ls:
		import numpy
		n = len(x_ls)
		xMedianValue =  numpy.median(x_ls)
		yMedianValue =  numpy.median(y_ls)
		corr = numpy.corrcoef(x_ls, y_ls)[0,1]
		wilcoxonSignedRankTestMinRankSumDif, wilcoxonSignedRankTestPvalue = stats.wilcoxon(x=x_ls, y=y_ls, zero_method='wilcox', correction=False)
		title = "n=%s p=%.4g r2=%.3f xMed=%.3g yMed=%.3g"%(n, wilcoxonSignedRankTestPvalue, 
															corr, xMedianValue, yMedianValue)
	else:
		title = ""
	return title

def constructTitleFromDataSummaryStat(data_ls=None):
	"""
	2012.8.17
	"""
	import numpy
	medianValue =  numpy.median(data_ls)
	meanValue = numpy.mean(data_ls)
	std = numpy.std(data_ls)
	title = 'n=%s mean=%.3g med=%.3g std %.3g'%(len(data_ls), meanValue, medianValue, std)
	return title

if __name__ == '__main__':
	#import pdb
	#pdb.set_trace()
	import pylab
	ax = pylab.gca()
	fig = pylab.gcf()
	name_ls = ['1', '2', '3']
	name2fc = assignMatPlotlibHueColorToLs(name_ls)
	drawName2FCLegend(ax, name_ls, name2fc, shape_type=1, no_face_color=True, no_edge_color=False, font_size=10)
	pylab.show()